# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_EmptyArray():
    a = ak.contents.emptyarray.EmptyArray()
    assert len(a) == 0
    with pytest.raises(IndexError):
        a[0,]
    assert isinstance(
        a[10:20,],
        ak.contents.emptyarray.EmptyArray,
    )
    assert len(a[10:20,]) == 0
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_NumpyArray_to_RegularArray():
    a = ak.operations.from_numpy(np.arange(2 * 3 * 5).reshape(2, 3, 5)).layout
    b = a.to_RegularArray()
    assert isinstance(b, ak.contents.RegularArray)
    assert len(b) == len(a) == 2
    assert b.size == 3
    assert isinstance(b.content, ak.contents.RegularArray)
    assert len(b.content) == 6
    assert b.content.size == 5
    assert isinstance(b.content.content, ak.contents.NumpyArray)
    assert len(b.content.content) == 30
    assert a.to_typetracer().to_RegularArray().form == b.form

    a = ak.operations.from_numpy(np.arange(2 * 0 * 5).reshape(2, 0, 5)).layout
    b = a.to_RegularArray()
    assert isinstance(b, ak.contents.RegularArray)
    assert len(b) == len(a) == 2
    assert b.size == 0
    assert isinstance(b.content, ak.contents.RegularArray)
    assert len(b.content) == 0
    assert b.content.size == 5
    assert isinstance(b.content.content, ak.contents.NumpyArray)
    assert len(b.content.content) == 0
    assert a.to_typetracer().to_RegularArray().form == b.form


def test_NumpyArray():
    a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
    )
    assert len(a) == 4
    assert a[2,] == 2.2
    assert a[-2,] == 2.2
    assert type(a[2,]) is np.float64
    with pytest.raises(IndexError):
        a[4,]
    with pytest.raises(IndexError):
        a[-5,]
    assert isinstance(
        a[2:,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a[2:,][0,] == 2.2
    assert len(a[2:,]) == 2
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]

    b = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )
    assert len(b) == 2
    assert isinstance(
        b[1,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert len(b[1,]) == 3
    with pytest.raises(IndexError):
        b[2,]
    assert b[1,][2,][0,] == 25
    assert len(b[1,][2,][1:,]) == 4
    assert b[1,][2,][1:,][2,] == 28
    with pytest.raises(IndexError):
        b["bad",]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"],]


def test_RegularArray_NumpyArray():
    # 6.6 is inaccessible
    a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
        ),
        3,
    )
    assert len(a) == 2
    assert isinstance(
        a[1,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()[1,].form == a[1,].form
    assert len(a[1,]) == 3
    assert a[1,][2,] == 5.5
    assert a[-1,][2,] == 5.5
    assert isinstance(
        a[1:2,],
        ak.contents.regulararray.RegularArray,
    )
    assert a.to_typetracer()[1:2,].form == a[1:2,].form
    assert len(a[1:,]) == 1
    assert len(a[1:100,]) == 1
    with pytest.raises(IndexError):
        a[2,]
    with pytest.raises(IndexError):
        a[-3,]
    with pytest.raises(IndexError):
        a[1,][3,]
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]

    b = ak.contents.regulararray.RegularArray(
        ak.contents.emptyarray.EmptyArray(), 0, zeros_length=10
    )
    assert len(b) == 10
    assert isinstance(
        b[5,],
        ak.contents.emptyarray.EmptyArray,
    )
    assert b.to_typetracer()[5,].form == b[5,].form
    assert len(b[5,]) == 0
    assert isinstance(
        b[7:,],
        ak.contents.regulararray.RegularArray,
    )
    assert b.to_typetracer()[7:,].form == b[7:,].form
    assert len(b[7:,]) == 3
    assert len(b[7:100,]) == 3
    with pytest.raises(IndexError):
        b["bad",]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"],]


def test_ListArray_NumpyArray():
    # 200 is inaccessible in stops
    # 6.6, 7.7, and 8.8 are inaccessible in content
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], dtype=np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )
    assert len(a) == 3
    with pytest.raises(IndexError):
        a[3,]
    with pytest.raises(IndexError):
        a[-4,]
    assert isinstance(
        a[2,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()[2,].form == a[2,].form
    assert len(a[0,]) == 3
    assert len(a[1,]) == 0
    assert len(a[2,]) == 2
    assert len(a[-3,]) == 3
    assert len(a[-2,]) == 0
    assert len(a[-1,]) == 2
    assert a[0,][-1,] == 3.3
    assert a[2,][-1,] == 5.5
    assert isinstance(
        a[1:,],
        ak.contents.listarray.ListArray,
    )
    assert a.to_typetracer()[1:,].form == a[1:,].form
    assert len(a[1:,]) == 2
    assert len(a[-2:,]) == 2
    assert len(a[1:100,]) == 2
    assert len(a[-2:100,]) == 2
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_ListOffsetArray_NumpyArray():
    # 6.6 and 7.7 are inaccessible
    a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6])),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
        ),
    )
    assert len(a) == 3
    with pytest.raises(IndexError):
        a[3,]
    with pytest.raises(IndexError):
        a[-4,]
    assert isinstance(
        a[2,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()[2,].form == a[2,].form
    assert len(a[0,]) == 3
    assert len(a[1,]) == 0
    assert len(a[2,]) == 2
    assert len(a[-3,]) == 3
    assert len(a[-2,]) == 0
    assert len(a[-1,]) == 2
    assert a[0,][-1,] == 3.3
    assert a[2,][-1,] == 5.5
    assert isinstance(
        a[1:,],
        ak.contents.listarray.ListArray,
    )
    assert a.to_typetracer()[1:,].form == a[1:,].form
    assert len(a[1:,]) == 2
    assert len(a[-2:,]) == 2
    assert len(a[1:100,]) == 2
    assert len(a[-2:100,]) == 2
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_RecordArray_NumpyArray():
    # 5.5 is inaccessible
    a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4])),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
    )
    assert len(a) == 5
    with pytest.raises(IndexError):
        a[5,]
    with pytest.raises(IndexError):
        a[-6,]
    assert isinstance(
        a[2,],
        ak.record.Record,
    )
    assert a.to_typetracer()[2,].array.form == a[2,].array.form
    assert a[2,]["y"] == 2.2
    assert a[-3,]["y"] == 2.2
    assert isinstance(
        a[2:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert a.to_typetracer()[2:,].form == a[2:,].form
    assert len(a[2:,]) == 3
    assert len(a[-3:,]) == 3
    assert len(a[2:100,]) == 3
    assert len(a[-3:100,]) == 3
    assert isinstance(
        a["y",],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()["y",].form == a["y",].form
    assert a["y",][2,] == 2.2
    assert a["y",][-3,] == 2.2
    with pytest.raises(IndexError):
        a["z",]
    with pytest.raises(IndexError):
        a[["x", "z"],]
    assert len(a[["x", "y"],]) == 5

    # 5.5 is inaccessible
    b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4])),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )
    assert len(b) == 5
    with pytest.raises(IndexError):
        b[5,]
    with pytest.raises(IndexError):
        b[-6,]
    assert isinstance(
        b[2,],
        ak.record.Record,
    )
    assert b.to_typetracer()[2,].array.form == b[2,].array.form
    assert b[2,]["1",] == 2.2
    assert b[-3,]["1",] == 2.2
    assert isinstance(
        b[2:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert b.to_typetracer()[2:,].form == b[2:,].form
    assert len(b[2:,]) == 3
    assert len(b[-3:,]) == 3
    assert len(b[2:100,]) == 3
    assert len(b[-3:100,]) == 3
    assert isinstance(
        b["1",],
        ak.contents.numpyarray.NumpyArray,
    )
    assert b.to_typetracer()["1",].form == b["1",].form
    assert b["1",][2,] == 2.2
    assert b["1",][-3,] == 2.2
    with pytest.raises(IndexError):
        a["2",]
    assert len(b[["0", "1"],]) == 5

    c = ak.contents.recordarray.RecordArray([], [], 10)
    assert len(c) == 10
    assert isinstance(
        c[5,],
        ak.record.Record,
    )
    assert c.to_typetracer()[5,].array.form == c[5,].array.form
    assert isinstance(
        c[7:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert c.to_typetracer()[7:,].form == c[7:,].form
    assert len(c[7:,]) == 3
    assert len(c[-3:,]) == 3
    with pytest.raises(IndexError):
        c["x",]

    d = ak.contents.recordarray.RecordArray([], None, 10)
    assert len(d) == 10
    assert isinstance(
        d[5,],
        ak.record.Record,
    )
    assert d.to_typetracer()[5,].array.form == d[5,].array.form
    assert isinstance(
        d[7:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert d.to_typetracer()[7:,].form == d[7:,].form
    assert len(d[7:,]) == 3
    assert len(d[-3:,]) == 3
    with pytest.raises(IndexError):
        d["0",]


def test_IndexedArray_NumpyArray():
    # 4.4 is inaccessible; 3.3 and 5.5 appear twice
    a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    assert len(a) == 7
    assert a[0,] == 3.3
    assert a[1,] == 3.3
    assert a[2,] == 1.1
    assert a[3,] == 2.2
    assert a[4,] == 5.5
    assert a[5,] == 6.6
    assert a[6,] == 5.5
    assert a[-7,] == 3.3
    assert a[-6,] == 3.3
    assert a[-5,] == 1.1
    assert a[-4,] == 2.2
    assert a[-3,] == 5.5
    assert a[-2,] == 6.6
    assert a[-1,] == 5.5
    with pytest.raises(IndexError):
        a[7,]
    with pytest.raises(IndexError):
        a[-8,]
    assert isinstance(
        a[3:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert a.to_typetracer()[3:,].form == a[3:,].form
    assert len(a[3:,]) == 4
    assert len(a[-4:,]) == 4
    assert len(a[3:100,]) == 4
    assert len(a[-4:100,]) == 4
    assert a[3:,][1,] == 5.5
    assert a[-4:,][1,] == 5.5
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_IndexedOptionArray_NumpyArray():
    # 1.1 and 4.4 are inaccessible; 3.3 appears twice
    a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    assert len(a) == 7
    assert a[0,] == 3.3
    assert a[1,] == 3.3
    assert a[2,] is None
    assert a[3,] == 2.2
    assert a[4,] is None
    assert a[5,] == 6.6
    assert a[6,] == 5.5
    assert a[-7,] == 3.3
    assert a[-6,] == 3.3
    assert a[-5,] is None
    assert a[-4,] == 2.2
    assert a[-3,] is None
    assert a[-2,] == 6.6
    assert a[-1,] == 5.5
    with pytest.raises(IndexError):
        a[7,]
    with pytest.raises(IndexError):
        a[-8,]
    assert isinstance(
        a[3:,],
        ak.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert a.to_typetracer()[3:,].form == a[3:,].form
    assert len(a[3:,]) == 4
    assert len(a[-4:,]) == 4
    assert len(a[3:100,]) == 4
    assert len(a[-4:100,]) == 4
    assert a[3:,][1,] is None
    assert a[-4:,][1,] is None
    assert a[3:,][2,] == 6.6
    assert a[-4:,][2,] == 6.6
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_ByteMaskedArray_NumpyArray():
    # 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], dtype=np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=True,
    )
    assert len(a) == 5
    with pytest.raises(IndexError):
        a[5,]
    with pytest.raises(IndexError):
        a[-6,]
    assert a[0,] == 1.1
    assert a[1,] is None
    assert a[2,] == 3.3
    assert a[3,] is None
    assert a[4,] == 5.5
    assert a[-5,] == 1.1
    assert a[-4,] is None
    assert a[-3,] == 3.3
    assert a[-2,] is None
    assert a[-1,] == 5.5
    assert isinstance(
        a[2:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert a.to_typetracer()[2:,].form == a[2:,].form
    assert len(a[2:,]) == 3
    assert len(a[-3:,]) == 3
    assert len(a[2:100,]) == 3
    assert len(a[-3:100,]) == 3
    assert a[2:,][1,] is None
    assert a[-3:,][1,] is None
    assert a[2:,][2,] == 5.5
    assert a[-3:,][2,] == 5.5
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]

    # 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=False,
    )
    assert len(b) == 5
    with pytest.raises(IndexError):
        b[5,]
    with pytest.raises(IndexError):
        b[-6,]
    assert b[0,] == 1.1
    assert b[1,] is None
    assert b[2,] == 3.3
    assert b[3,] is None
    assert b[4,] == 5.5
    assert b[-5,] == 1.1
    assert b[-4,] is None
    assert b[-3,] == 3.3
    assert b[-2,] is None
    assert b[-1,] == 5.5
    assert isinstance(
        b[2:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert b.to_typetracer()[2:,].form == b[2:,].form
    assert len(b[2:,]) == 3
    assert len(b[-3:,]) == 3
    assert len(b[2:100,]) == 3
    assert len(b[-3:100,]) == 3
    assert b[2:,][1,] is None
    assert b[-3:,][1,] is None
    assert b[2:,][2,] == 5.5
    assert b[-3:,][2,] == 5.5
    with pytest.raises(IndexError):
        b["bad",]
    with pytest.raises(IndexError):
        b[["bad", "good", "ok"],]


def test_BitMaskedArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    assert len(a) == 13
    with pytest.raises(IndexError):
        a[13,]
    with pytest.raises(IndexError):
        a[-14,]
    assert a[0,] == 0.0
    assert a[1,] == 1.0
    assert a[2,] == 2.0
    assert a[3,] == 3.0
    assert a[4,] is None
    assert a[5,] is None
    assert a[6,] is None
    assert a[7,] is None
    assert a[8,] == 1.1
    assert a[9,] is None
    assert a[10,] == 3.3
    assert a[11,] is None
    assert a[12,] == 5.5
    assert a[-13,] == 0.0
    assert a[-12,] == 1.0
    assert a[-11,] == 2.0
    assert a[-10,] == 3.0
    assert a[-9,] is None
    assert a[-8,] is None
    assert a[-7,] is None
    assert a[-6,] is None
    assert a[-5,] == 1.1
    assert a[-4,] is None
    assert a[-3,] == 3.3
    assert a[-2,] is None
    assert a[-1,] == 5.5

    assert isinstance(
        a[5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert a.to_typetracer()[5:,].form == a[5:,].form
    assert len(a[5:,]) == 8
    assert len(a[-8:,]) == 8
    assert len(a[5:100,]) == 8
    assert len(a[-8:100,]) == 8
    assert a[5:,][2,] is None
    assert a[5:,][3,] == 1.1
    assert a[-8:,][2,] is None
    assert a[-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        a["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    assert len(b) == 13
    with pytest.raises(IndexError):
        b[13,]
    with pytest.raises(IndexError):
        b[-14,]
    assert b[0,] == 0.0
    assert b[1,] == 1.0
    assert b[2,] == 2.0
    assert b[3,] == 3.0
    assert b[4,] is None
    assert b[5,] is None
    assert b[6,] is None
    assert b[7,] is None
    assert b[8,] == 1.1
    assert b[9,] is None
    assert b[10,] == 3.3
    assert b[11,] is None
    assert b[12,] == 5.5
    assert b[-13,] == 0.0
    assert b[-12,] == 1.0
    assert b[-11,] == 2.0
    assert b[-10,] == 3.0
    assert b[-9,] is None
    assert b[-8,] is None
    assert b[-7,] is None
    assert b[-6,] is None
    assert b[-5,] == 1.1
    assert b[-4,] is None
    assert b[-3,] == 3.3
    assert b[-2,] is None
    assert b[-1,] == 5.5
    assert isinstance(
        b[5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert b.to_typetracer()[5:,].form == b[5:,].form
    assert len(b[5:,]) == 8
    assert len(b[-8:,]) == 8
    assert len(b[5:100,]) == 8
    assert len(b[-8:100,]) == 8
    assert b[5:,][2,] is None
    assert b[5:,][3,] == 1.1
    assert b[-8:,][2,] is None
    assert b[-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        b["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    assert len(c) == 13
    with pytest.raises(IndexError):
        c[13,]
    with pytest.raises(IndexError):
        c[-14,]
    assert c[0,] == 0.0
    assert c[1,] == 1.0
    assert c[2,] == 2.0
    assert c[3,] == 3.0
    assert c[4,] is None
    assert c[5,] is None
    assert c[6,] is None
    assert c[7,] is None
    assert c[8,] == 1.1
    assert c[9,] is None
    assert c[10,] == 3.3
    assert c[11,] is None
    assert c[12,] == 5.5
    assert c[-13,] == 0.0
    assert c[-12,] == 1.0
    assert c[-11,] == 2.0
    assert c[-10,] == 3.0
    assert c[-9,] is None
    assert c[-8,] is None
    assert c[-7,] is None
    assert c[-6,] is None
    assert c[-5,] == 1.1
    assert c[-4,] is None
    assert c[-3,] == 3.3
    assert c[-2,] is None
    assert c[-1,] == 5.5
    assert isinstance(
        c[5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert c.to_typetracer()[5:,].form == c[5:,].form
    assert len(c[5:,]) == 8
    assert len(c[-8:,]) == 8
    assert len(c[5:100,]) == 8
    assert len(c[-8:100,]) == 8
    assert c[5:,][2,] is None
    assert c[5:,][3,] == 1.1
    assert c[-8:,][2,] is None
    assert c[-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        c["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    assert len(d) == 13
    with pytest.raises(IndexError):
        d[13,]
    with pytest.raises(IndexError):
        d[-14,]
    assert d[0,] == 0.0
    assert d[1,] == 1.0
    assert d[2,] == 2.0
    assert d[3,] == 3.0
    assert d[4,] is None
    assert d[5,] is None
    assert d[6,] is None
    assert d[7,] is None
    assert d[8,] == 1.1
    assert d[9,] is None
    assert d[10,] == 3.3
    assert d[11,] is None
    assert d[12,] == 5.5
    assert d[-13,] == 0.0
    assert d[-12,] == 1.0
    assert d[-11,] == 2.0
    assert d[-10,] == 3.0
    assert d[-9,] is None
    assert d[-8,] is None
    assert d[-7,] is None
    assert d[-6,] is None
    assert d[-5,] == 1.1
    assert d[-4,] is None
    assert d[-3,] == 3.3
    assert d[-2,] is None
    assert d[-1,] == 5.5
    assert isinstance(
        d[5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert d.to_typetracer()[5:,].form == d[5:,].form
    assert len(d[5:,]) == 8
    assert len(d[-8:,]) == 8
    assert len(d[5:100,]) == 8
    assert len(d[-8:100,]) == 8
    assert d[5:,][2,] is None
    assert d[5:,][3,] == 1.1
    assert d[-8:,][2,] is None
    assert d[-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        d["bad",]


def test_UnmaskedArray_NumpyArray():
    a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
        )
    )
    assert len(a) == 4
    assert a[2,] == 2.2
    assert a[-2,] == 2.2
    assert type(a[2,]) is np.float64
    with pytest.raises(IndexError):
        a[4,]
    with pytest.raises(IndexError):
        a[-5,]
    assert isinstance(
        a[2:,],
        ak.contents.unmaskedarray.UnmaskedArray,
    )
    assert a.to_typetracer()[2:,].form == a[2:,].form
    assert a[2:,][0,] == 2.2
    assert len(a[2:,]) == 2
    with pytest.raises(IndexError):
        a["bad",]


def test_UnionArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3])),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    assert len(a) == 7
    with pytest.raises(IndexError):
        a[7,]
    with pytest.raises(IndexError):
        a[-8,]
    assert a[0,] == 5.5
    assert a[1,] == 4.4
    assert a[2,] == 1.0
    assert a[3,] == 2.0
    assert a[4,] == 3.3
    assert a[5,] == 3.0
    assert a[6,] == 5.5
    assert a[-7,] == 5.5
    assert a[-6,] == 4.4
    assert a[-5,] == 1.0
    assert a[-4,] == 2.0
    assert a[-3,] == 3.3
    assert a[-2,] == 3.0
    assert a[-1,] == 5.5
    assert a.to_typetracer()[3:,].form == a[3:,].form
    assert len(a[3:,]) == 4
    assert len(a[-4:,]) == 4
    assert len(a[3:100,]) == 4
    assert len(a[-4:100,]) == 4
    assert a[3:,][1,] == 3.3
    assert a[3:,][2,] == 3.0
    assert a[-4:,][1,] == 3.3
    assert a[-4:,][2,] == 3.0
    with pytest.raises(IndexError):
        a["bad",]
    with pytest.raises(IndexError):
        a[["bad", "good", "ok"],]


def test_RegularArray_RecordArray_NumpyArray():
    # 6.6 is inaccessible
    a = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        3,
    )
    assert len(a["nest",]) == 2
    assert isinstance(
        a["nest",][1,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()["nest",][1,].form == a["nest",][1,].form
    assert len(a["nest",][1,]) == 3
    assert a["nest",][1,][2,] == 5.5
    assert a["nest",][-1,][2,] == 5.5
    assert isinstance(
        a["nest",][1:2,],
        ak.contents.regulararray.RegularArray,
    )
    assert a.to_typetracer()["nest",][1:2,].form == a["nest",][1:2,].form
    assert len(a["nest",][1:,]) == 1
    assert len(a["nest",][1:100,]) == 1
    with pytest.raises(IndexError):
        a["nest",][2,]
    with pytest.raises(IndexError):
        a["nest",][-3,]
    with pytest.raises(IndexError):
        a["nest",][1,][3,]
    with pytest.raises(IndexError):
        a["nest",]["bad",]

    b = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [ak.contents.emptyarray.EmptyArray()], ["nest"]
        ),
        0,
        zeros_length=10,
    )
    assert len(b["nest",]) == 10
    assert isinstance(
        b["nest",][5,],
        ak.contents.emptyarray.EmptyArray,
    )
    assert b.to_typetracer()["nest",][5,].form == b["nest",][5,].form
    assert len(b["nest",][5,]) == 0
    assert isinstance(
        b["nest",][7:,],
        ak.contents.regulararray.RegularArray,
    )
    assert b.to_typetracer()["nest",][7:,].form == b["nest",][7:,].form
    assert len(b["nest",][7:,]) == 3
    assert len(b["nest",][7:100,]) == 3
    with pytest.raises(IndexError):
        b["nest",]["bad",]


def test_ListArray_RecordArray_NumpyArray():
    # 200 is inaccessible in stops
    # 6.6, 7.7, and 8.8 are inaccessible in content
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1])),
        ak.index.Index(np.array([7, 100, 3, 200])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
                )
            ],
            ["nest"],
        ),
    )
    assert len(a["nest",]) == 3
    with pytest.raises(IndexError):
        a["nest",][3,]
    with pytest.raises(IndexError):
        a["nest",][-4,]
    assert isinstance(
        a["nest",][2,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()["nest",][2,].form == a["nest",][2,].form
    assert len(a["nest",][0,]) == 3
    assert len(a["nest",][1,]) == 0
    assert len(a["nest",][2,]) == 2
    assert len(a["nest",][-3,]) == 3
    assert len(a["nest",][-2,]) == 0
    assert len(a["nest",][-1,]) == 2
    assert a["nest",][0,][-1,] == 3.3
    assert a["nest",][2,][-1,] == 5.5
    assert isinstance(
        a["nest",][1:,],
        ak.contents.listarray.ListArray,
    )
    assert len(a["nest",][1:,]) == 2
    assert len(a["nest",][-2:,]) == 2
    assert len(a["nest",][1:100,]) == 2
    assert len(a["nest",][-2:100,]) == 2
    with pytest.raises(IndexError):
        a["nest",]["bad",]


def test_ListOffsetArray_RecordArray_NumpyArray():
    # 6.6 and 7.7 are inaccessible
    a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
                )
            ],
            ["nest"],
        ),
    )
    assert len(a["nest",]) == 3
    with pytest.raises(IndexError):
        a["nest",][3,]
    with pytest.raises(IndexError):
        a["nest",][-4,]
    assert isinstance(
        a["nest",][2,],
        ak.contents.numpyarray.NumpyArray,
    )
    assert a.to_typetracer()["nest",][2,].form == a["nest",][2,].form
    assert len(a["nest",][0,]) == 3
    assert len(a["nest",][1,]) == 0
    assert len(a["nest",][2,]) == 2
    assert len(a["nest",][-3,]) == 3
    assert len(a["nest",][-2,]) == 0
    assert len(a["nest",][-1,]) == 2
    assert a["nest",][0,][-1,] == 3.3
    assert a["nest",][2,][-1,] == 5.5
    assert isinstance(
        a["nest",][1:,],
        ak.contents.listarray.ListArray,
    )
    assert a.to_typetracer()["nest",][1:,].form == a["nest",][1:,].form
    assert len(a["nest",][1:,]) == 2
    assert len(a["nest",][-2:,]) == 2
    assert len(a["nest",][1:100,]) == 2
    assert len(a["nest",][-2:100,]) == 2
    with pytest.raises(IndexError):
        a["nest",]["bad",]


def test_IndexedArray_RecordArray_NumpyArray():
    # 4.4 is inaccessible; 3.3 and 5.5 appear twice
    a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert len(a["nest",]) == 7
    assert a["nest",][0,] == 3.3
    assert a["nest",][1,] == 3.3
    assert a["nest",][2,] == 1.1
    assert a["nest",][3,] == 2.2
    assert a["nest",][4,] == 5.5
    assert a["nest",][5,] == 6.6
    assert a["nest",][6,] == 5.5
    assert a["nest",][-7,] == 3.3
    assert a["nest",][-6,] == 3.3
    assert a["nest",][-5,] == 1.1
    assert a["nest",][-4,] == 2.2
    assert a["nest",][-3,] == 5.5
    assert a["nest",][-2,] == 6.6
    assert a["nest",][-1,] == 5.5
    with pytest.raises(IndexError):
        a["nest",][7,]
    with pytest.raises(IndexError):
        a["nest",][-8,]
    assert isinstance(
        a["nest",][3:,],
        ak.contents.indexedarray.IndexedArray,
    )
    assert a.to_typetracer()["nest",][3:,].form == a["nest",][3:,].form
    assert len(a["nest",][3:,]) == 4
    assert len(a["nest",][-4:,]) == 4
    assert len(a["nest",][3:100,]) == 4
    assert len(a["nest",][-4:100,]) == 4
    assert a["nest",][3:,][1,] == 5.5
    assert a["nest",][-4:,][1,] == 5.5
    with pytest.raises(IndexError):
        a["nest",]["bad",]


def test_IndexedOptionArray_RecordArray_NumpyArray():
    # 1.1 and 4.4 are inaccessible; 3.3 appears twice
    a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert len(a["nest",]) == 7
    assert a["nest",][0,] == 3.3
    assert a["nest",][1,] == 3.3
    assert a["nest",][2,] is None
    assert a["nest",][3,] == 2.2
    assert a["nest",][4,] is None
    assert a["nest",][5,] == 6.6
    assert a["nest",][6,] == 5.5
    assert a["nest",][-7,] == 3.3
    assert a["nest",][-6,] == 3.3
    assert a["nest",][-5,] is None
    assert a["nest",][-4,] == 2.2
    assert a["nest",][-3,] is None
    assert a["nest",][-2,] == 6.6
    assert a["nest",][-1,] == 5.5
    with pytest.raises(IndexError):
        a["nest",][7,]
    with pytest.raises(IndexError):
        a["nest",][-8,]
    assert isinstance(
        a["nest",][3:,],
        ak.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert a.to_typetracer()["nest",][3:,].form == a["nest",][3:,].form
    assert len(a["nest",][3:,]) == 4
    assert len(a["nest",][-4:,]) == 4
    assert len(a["nest",][3:100,]) == 4
    assert len(a["nest",][-4:100,]) == 4
    assert a["nest",][3:,][1,] is None
    assert a["nest",][-4:,][1,] is None
    assert a["nest",][3:,][2,] == 6.6
    assert a["nest",][-4:,][2,] == 6.6
    with pytest.raises(IndexError):
        a["nest",]["bad",]


def test_ByteMaskedArray_RecordArray_NumpyArray():
    # 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=True,
    )
    assert len(a["nest",]) == 5
    with pytest.raises(IndexError):
        a["nest",][5,]
    with pytest.raises(IndexError):
        a["nest",][-6,]
    assert a["nest",][0,] == 1.1
    assert a["nest",][1,] is None
    assert a["nest",][2,] == 3.3
    assert a["nest",][3,] is None
    assert a["nest",][4,] == 5.5
    assert a["nest",][-5,] == 1.1
    assert a["nest",][-4,] is None
    assert a["nest",][-3,] == 3.3
    assert a["nest",][-2,] is None
    assert a["nest",][-1,] == 5.5
    assert isinstance(
        a["nest",][2:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert a.to_typetracer()["nest",][2:,].form == a["nest",][2:,].form
    assert len(a["nest",][2:,]) == 3
    assert len(a["nest",][-3:,]) == 3
    assert len(a["nest",][2:100,]) == 3
    assert len(a["nest",][-3:100,]) == 3
    assert a["nest",][2:,][1,] is None
    assert a["nest",][-3:,][1,] is None
    assert a["nest",][2:,][2,] == 5.5
    assert a["nest",][-3:,][2,] == 5.5
    with pytest.raises(IndexError):
        a["nest",]["bad",]

    # 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=False,
    )
    assert len(b["nest",]) == 5
    with pytest.raises(IndexError):
        b["nest",][5,]
    with pytest.raises(IndexError):
        b["nest",][-6,]
    assert b["nest",][0,] == 1.1
    assert b["nest",][1,] is None
    assert b["nest",][2,] == 3.3
    assert b["nest",][3,] is None
    assert b["nest",][4,] == 5.5
    assert b["nest",][-5,] == 1.1
    assert b["nest",][-4,] is None
    assert b["nest",][-3,] == 3.3
    assert b["nest",][-2,] is None
    assert b["nest",][-1,] == 5.5
    assert isinstance(
        b["nest",][2:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert b.to_typetracer()["nest",][2:,].form == b["nest",][2:,].form
    assert len(b["nest",][2:,]) == 3
    assert len(b["nest",][-3:,]) == 3
    assert len(b["nest",][2:100,]) == 3
    assert len(b["nest",][-3:100,]) == 3
    assert b["nest",][2:,][1,] is None
    assert b["nest",][-3:,][1,] is None
    assert b["nest",][2:,][2,] == 5.5
    assert b["nest",][-3:,][2,] == 5.5
    with pytest.raises(IndexError):
        b["nest",]["bad",]


def test_BitMaskedArray_RecordArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        True,
                        True,
                        True,
                        True,
                        False,
                        False,
                        False,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                    ]
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    assert len(a["nest"]) == 13
    with pytest.raises(IndexError):
        a["nest",][13,]
    with pytest.raises(IndexError):
        a["nest",][-14,]
    assert a["nest",][0,] == 0.0
    assert a["nest",][1,] == 1.0
    assert a["nest",][2,] == 2.0
    assert a["nest",][3,] == 3.0
    assert a["nest",][4,] is None
    assert a["nest",][5,] is None
    assert a["nest",][6,] is None
    assert a["nest",][7,] is None
    assert a["nest",][8,] == 1.1
    assert a["nest",][9,] is None
    assert a["nest",][10,] == 3.3
    assert a["nest",][11,] is None
    assert a["nest",][12,] == 5.5
    assert a["nest",][-13,] == 0.0
    assert a["nest",][-12,] == 1.0
    assert a["nest",][-11,] == 2.0
    assert a["nest",][-10,] == 3.0
    assert a["nest",][-9,] is None
    assert a["nest",][-8,] is None
    assert a["nest",][-7,] is None
    assert a["nest",][-6,] is None
    assert a["nest",][-5,] == 1.1
    assert a["nest",][-4,] is None
    assert a["nest",][-3,] == 3.3
    assert a["nest",][-2,] is None
    assert a["nest",][-1,] == 5.5
    assert isinstance(
        a["nest",][5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert a.to_typetracer()["nest",][5:,].form == a["nest",][5:,].form
    assert len(a["nest",][5:,]) == 8
    assert len(a["nest",][-8:,]) == 8
    assert len(a["nest",][5:100,]) == 8
    assert len(a["nest",][-8:100,]) == 8
    assert a["nest",][5:,][2,] is None
    assert a["nest",][5:,][3,] == 1.1
    assert a["nest",][-8:,][2,] is None
    assert a["nest",][-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        a["nest",]["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    assert len(b["nest",]) == 13
    with pytest.raises(IndexError):
        b["nest",][13,]
    with pytest.raises(IndexError):
        b["nest",][-14,]
    assert b["nest",][0,] == 0.0
    assert b["nest",][1,] == 1.0
    assert b["nest",][2,] == 2.0
    assert b["nest",][3,] == 3.0
    assert b["nest",][4,] is None
    assert b["nest",][5,] is None
    assert b["nest",][6,] is None
    assert b["nest",][7,] is None
    assert b["nest",][8,] == 1.1
    assert b["nest",][9,] is None
    assert b["nest",][10,] == 3.3
    assert b["nest",][11,] is None
    assert b["nest",][12,] == 5.5
    assert b["nest",][-13,] == 0.0
    assert b["nest",][-12,] == 1.0
    assert b["nest",][-11,] == 2.0
    assert b["nest",][-10,] == 3.0
    assert b["nest",][-9,] is None
    assert b["nest",][-8,] is None
    assert b["nest",][-7,] is None
    assert b["nest",][-6,] is None
    assert b["nest",][-5,] == 1.1
    assert b["nest",][-4,] is None
    assert b["nest",][-3,] == 3.3
    assert b["nest",][-2,] is None
    assert b["nest",][-1,] == 5.5
    assert isinstance(
        b["nest",][5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert b.to_typetracer()["nest",][5:,].form == b["nest",][5:,].form
    assert len(b["nest",][5:,]) == 8
    assert len(b["nest",][-8:,]) == 8
    assert len(b["nest",][5:100,]) == 8
    assert len(b["nest",][-8:100,]) == 8
    assert b["nest",][5:,][2,] is None
    assert b["nest",][5:,][3,] == 1.1
    assert b["nest",][-8:,][2,] is None
    assert b["nest",][-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        b["nest",]["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    assert len(c["nest",]) == 13
    with pytest.raises(IndexError):
        c["nest",][13,]
    with pytest.raises(IndexError):
        c["nest",][-14,]
    assert c["nest",][0,] == 0.0
    assert c["nest",][1,] == 1.0
    assert c["nest",][2,] == 2.0
    assert c["nest",][3,] == 3.0
    assert c["nest",][4,] is None
    assert c["nest",][5,] is None
    assert c["nest",][6,] is None
    assert c["nest",][7,] is None
    assert c["nest",][8,] == 1.1
    assert c["nest",][9,] is None
    assert c["nest",][10,] == 3.3
    assert c["nest",][11,] is None
    assert c["nest",][12,] == 5.5
    assert c["nest",][-13,] == 0.0
    assert c["nest",][-12,] == 1.0
    assert c["nest",][-11,] == 2.0
    assert c["nest",][-10,] == 3.0
    assert c["nest",][-9,] is None
    assert c["nest",][-8,] is None
    assert c["nest",][-7,] is None
    assert c["nest",][-6,] is None
    assert c["nest",][-5,] == 1.1
    assert c["nest",][-4,] is None
    assert c["nest",][-3,] == 3.3
    assert c["nest",][-2,] is None
    assert c["nest",][-1,] == 5.5
    assert isinstance(
        c["nest",][5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert c.to_typetracer()["nest",][5:,].form == c["nest",][5:,].form
    assert len(c["nest",][5:,]) == 8
    assert len(c["nest",][-8:,]) == 8
    assert len(c["nest",][5:100,]) == 8
    assert len(c["nest",][-8:100,]) == 8
    assert c["nest",][5:,][2,] is None
    assert c["nest",][5:,][3,] == 1.1
    assert c["nest",][-8:,][2,] is None
    assert c["nest",][-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        c["nest",]["bad",]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    assert len(d["nest",]) == 13
    with pytest.raises(IndexError):
        d["nest",][13,]
    with pytest.raises(IndexError):
        d["nest",][-14,]
    assert d["nest",][0,] == 0.0
    assert d["nest",][1,] == 1.0
    assert d["nest",][2,] == 2.0
    assert d["nest",][3,] == 3.0
    assert d["nest",][4,] is None
    assert d["nest",][5,] is None
    assert d["nest",][6,] is None
    assert d["nest",][7,] is None
    assert d["nest",][8,] == 1.1
    assert d["nest",][9,] is None
    assert d["nest",][10,] == 3.3
    assert d["nest",][11,] is None
    assert d["nest",][12,] == 5.5
    assert d["nest",][-13,] == 0.0
    assert d["nest",][-12,] == 1.0
    assert d["nest",][-11,] == 2.0
    assert d["nest",][-10,] == 3.0
    assert d["nest",][-9,] is None
    assert d["nest",][-8,] is None
    assert d["nest",][-7,] is None
    assert d["nest",][-6,] is None
    assert d["nest",][-5,] == 1.1
    assert d["nest",][-4,] is None
    assert d["nest",][-3,] == 3.3
    assert d["nest",][-2,] is None
    assert d["nest",][-1,] == 5.5
    assert isinstance(
        d["nest",][5:,],
        ak.contents.bytemaskedarray.ByteMaskedArray,
    )
    assert d.to_typetracer()["nest",][5:,].form == d["nest",][5:,].form
    assert len(d["nest",][5:,]) == 8
    assert len(d["nest",][-8:,]) == 8
    assert len(d["nest",][5:100,]) == 8
    assert len(d["nest",][-8:100,]) == 8
    assert d["nest",][5:,][2,] is None
    assert d["nest",][5:,][3,] == 1.1
    assert d["nest",][-8:,][2,] is None
    assert d["nest",][-8:,][3,] == 1.1
    with pytest.raises(IndexError):
        d["nest",]["bad",]


def test_UnmaskedArray_RecordArray_NumpyArray():
    a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
                )
            ],
            ["nest"],
        )
    )
    assert len(a["nest",]) == 4
    assert a["nest",][2,] == 2.2
    assert a["nest",][-2,] == 2.2
    assert type(a["nest",][2,]) is np.float64
    with pytest.raises(IndexError):
        a["nest",][4,]
    with pytest.raises(IndexError):
        a["nest",][-5,]
    assert isinstance(
        a["nest",][2:,],
        ak.contents.unmaskedarray.UnmaskedArray,
    )
    assert a.to_typetracer()["nest",][2:,].form == a["nest",][2:,].form
    assert a["nest",][2:,][0,] == 2.2
    assert len(a["nest",][2:,]) == 2
    with pytest.raises(IndexError):
        a["nest",]["bad",]


def test_UnionArray_RecordArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.recordarray.RecordArray(
                [ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3]))], ["nest"]
            ),
            ak.contents.recordarray.RecordArray(
                [
                    ak.contents.numpyarray.NumpyArray(
                        np.array([1.1, 2.2, 3.3, 4.4, 5.5])
                    )
                ],
                ["nest"],
            ),
        ],
    )
    assert len(a["nest",]) == 7
    with pytest.raises(IndexError):
        a["nest",][7,]
    with pytest.raises(IndexError):
        a["nest",][-8,]
    assert a["nest",][0,] == 5.5
    assert a["nest",][1,] == 4.4
    assert a["nest",][2,] == 1.0
    assert a["nest",][3,] == 2.0
    assert a["nest",][4,] == 3.3
    assert a["nest",][5,] == 3.0
    assert a["nest",][6,] == 5.5
    assert a["nest",][-7,] == 5.5
    assert a["nest",][-6,] == 4.4
    assert a["nest",][-5,] == 1.0
    assert a["nest",][-4,] == 2.0
    assert a["nest",][-3,] == 3.3
    assert a["nest",][-2,] == 3.0
    assert a["nest",][-1,] == 5.5
    assert a.to_typetracer()["nest",][3:,].form == a["nest",][3:,].form
    assert len(a["nest",][3:,]) == 4
    assert len(a["nest",][-4:,]) == 4
    assert len(a["nest",][3:100,]) == 4
    assert len(a["nest",][-4:100,]) == 4
    assert a["nest",][3:,][1,] == 3.3
    assert a["nest",][3:,][2,] == 3.0
    assert a["nest",][-4:,][1,] == 3.3
    assert a["nest",][-4:,][2,] == 3.0
    with pytest.raises(IndexError):
        a["nest",]["bad",]



from __future__ import annotations

import awkward._nplikes.cupy
import awkward._nplikes.jax
import awkward._nplikes.numpy
import awkward._nplikes.typetracer
from awkward._nplikes.dispatch import nplike_of
from awkward._typing import TYPE_CHECKING

if TYPE_CHECKING:
    from awkward._nplikes.numpylike import ArrayLike, NumpyLike


def to_nplike(
    array: ArrayLike, nplike: NumpyLike, *, from_nplike: NumpyLike | None = None
) -> ArrayLike:
    if from_nplike is None:
        from_nplike = nplike_of(array, default=None)
        if from_nplike is None:
            raise TypeError(
                f"internal error: expected an array supported by an existing nplike, got {type(array).__name__!r}"
            )

    if from_nplike is to_nplike:
        return array

    if nplike.known_data and not from_nplike.known_data:
        raise TypeError(
            "Converting from an nplike without known data to an nplike with known data is not supported"
        )

    # Copy to host memory
    if isinstance(from_nplike, awkward._nplikes.cupy.Cupy):
        array = array.get()

    return nplike.asarray(array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array([[[], [], []], []])
    flattened = ak.flatten(array, axis=None)
    assert isinstance(flattened.layout, ak.contents.EmptyArray)
    assert len(flattened) == 0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import numpy as np

import awkward as ak


def test():
    layout = ak.contents.ByteMaskedArray(
        ak.index.Index8([0]),
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 1]),
            ak.contents.ByteMaskedArray(
                ak.index.Index8([1]),
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 2]), ak.contents.NumpyArray([5, 6])
                ),
                valid_when=True,
            ),
        ),
        valid_when=True,
    )
    assert ak.almost_equal(
        ak.any(layout, axis=2),
        ak.contents.IndexedOptionArray(
            ak.index.Index64([-1]),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0]),
                ak.contents.IndexedOptionArray(
                    ak.index.Index64([]),
                    ak.contents.NumpyArray(np.empty(0, dtype=np.bool_)),
                ),
            ),
        ),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array([[1, 2], [], [-np.inf]])

    # monoidal reducers (have identities)
    assert ak.count(array, axis=1)[1] == 0
    assert ak.count_nonzero(array, axis=1)[1] == 0
    assert ak.sum(array, axis=1)[1] == 0
    assert ak.prod(array, axis=1)[1] == 1
    assert ak.all(array, axis=1)[1]
    assert not ak.any(array, axis=1)[1]

    # semigroup reducers (no identity; arguable in the case of min/max on floats)
    assert ak.max(array, axis=1)[1] is None
    assert ak.min(array, axis=1)[1] is None
    assert ak.argmax(array, axis=1)[1] is None
    assert ak.argmin(array, axis=1)[1] is None

    # defined in terms of reducers, with ak.count in the denominator
    assert np.isnan(ak.moment(array, 1, axis=1)[1])
    assert np.isnan(ak.mean(array, axis=1)[1])
    assert np.isnan(ak.var(array, axis=1)[1])
    assert np.isnan(ak.std(array, axis=1)[1])
    assert np.isnan(ak.corr(array, array, axis=1)[1])
    assert np.isnan(ak.covar(array, array, axis=1)[1])

    assert np.isnan(ak.linear_fit(array, array, axis=1)["intercept", 1])
    assert np.isnan(ak.linear_fit(array, array, axis=1)["slope", 1])
    assert np.isnan(ak.linear_fit(array, array, axis=1)["intercept_error", 1])
    assert np.isnan(ak.linear_fit(array, array, axis=1)["slope_error", 1])

    # defined in terms of reducers, but a map-like operation
    assert ak.softmax(array, axis=1)[1].to_list() == []
    assert np.isnan(ak.softmax(array, axis=1)[2, 0])

    # defined in terms of reducers, computed from ak.min and ak.max
    assert ak.ptp(array, axis=1)[1] is None



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.from_numpy(np.arange(2 * 3 * 4 * 5, dtype=np.int64).reshape(2, 3, 4, 5))
    assert array.type == ak.types.ArrayType(
        ak.types.RegularType(
            ak.types.RegularType(
                ak.types.RegularType(ak.types.NumpyType("int64"), 5), 4
            ),
            3,
        ),
        2,
    )

    def apply(layout, **kwargs):
        return

    result_regular = ak.transform(
        apply,
        array,
        numpy_to_regular=True,
    )
    assert result_regular.type == ak.types.ArrayType(
        ak.types.RegularType(
            ak.types.RegularType(
                ak.types.RegularType(ak.types.NumpyType("int64"), 5), 4
            ),
            3,
        ),
        2,
    )

    result_ragged = ak.transform(
        apply,
        array,
        regular_to_jagged=True,
        numpy_to_regular=True,
    )
    assert result_ragged.type == ak.types.ArrayType(
        ak.types.ListType(
            ak.types.ListType(ak.types.ListType(ak.types.NumpyType("int64")))
        ),
        2,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_numpy_array():
    x = np.arange(12)
    y = ak.operations.is_none(x)

    assert y.to_list() == [False] * 12


def test_awkward_from_numpy_array():
    x = np.arange(12)
    y = ak.operations.from_numpy(x)
    z = ak.operations.is_none(y)

    assert z.to_list() == [False] * 12



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak
import awkward._connect.cling
import awkward._lookup

ROOT = pytest.importorskip("ROOT")


compiler = ROOT.gInterpreter.Declare


def test_NumpyArray_shape():
    ak_array_in = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert (
        data_frame.GetColumnType("x")
        == "ROOT::VecOps::RVec<ROOT::VecOps::RVec<int64_t>>"
    )


def test_RegularArray_NumpyArray():
    ak_array_in = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        3,
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<double>"


def test_ListArray_NumpyArray():
    ak_array_in = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<double>"


def test_ListOffsetArray_NumpyArray():
    ak_array_in = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6, 7], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<double>"


def test_RecordArray_NumpyArray():
    ak_array_in_one = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
        parameters={"__record__": "Something"},
    )
    data_frame_one = ak.to_rdataframe({"one": ak_array_in_one})
    assert str(data_frame_one.GetColumnType("one")).startswith(
        "awkward::Record_Something_"
    )


def test_RecordArray_NumpyArray_two():
    ak_array_two = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert str(data_frame_two.GetColumnType("two")).startswith("awkward::Record_")


def test_RecordArray_NumpyArray_three():
    ak_array_three = ak.contents.recordarray.RecordArray([], [], 10)
    data_frame_three = ak.to_rdataframe({"three": ak_array_three})
    assert str(data_frame_three.GetColumnType("three")).startswith("awkward::Record_")


def test_RecordArray_NumpyArray_four():
    ak_array_four = ak.contents.recordarray.RecordArray([], None, 10)
    data_frame_four = ak.to_rdataframe({"four": ak_array_four})
    assert str(data_frame_four.GetColumnType("four")).startswith("awkward::Record_")


def test_IndexedArray_NumpyArray():
    ak_array_in = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "double"


def test_IndexedOptionArray_NumpyArray():
    ak_array_in = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "std::optional<double>"


def test_ByteMaskedArray_NumpyArray():
    ak_array_one = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=True,
    )
    data_frame_one = ak.to_rdataframe({"one": ak_array_one})
    assert data_frame_one.GetColumnType("one") == "std::optional<double>"


def test_ByteMaskedArray_NumpyArray_two():
    ak_array_two = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=False,
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert data_frame_two.GetColumnType("two") == "std::optional<double>"


def test_BitMaskedArray_NumpyArray():
    ak_array_one = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    data_frame_one = ak.to_rdataframe({"one": ak_array_one})
    assert data_frame_one.GetColumnType("one") == "std::optional<double>"


def test_BitMaskedArray_NumpyArray_two():
    ak_array_two = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert data_frame_two.GetColumnType("two") == "std::optional<double>"


def test_BitMaskedArray_NumpyArray_three():
    ak_array_three = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    data_frame_three = ak.to_rdataframe({"three": ak_array_three})
    assert data_frame_three.GetColumnType("three") == "std::optional<double>"


def test_BitMaskedArray_NumpyArray_four():
    ak_array_four = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    data_frame_four = ak.to_rdataframe({"four": ak_array_four})
    assert data_frame_four.GetColumnType("four") == "std::optional<double>"


def test_UnmaskedArray_NumpyArray():
    ak_array_in = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3]))
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "std::optional<double>"


def test_nested_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([999.0, 0.0, 1.1, 2.2, 3.3]),
            parameters={"some": "stuff", "other": [1, 2, "three"]},
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<double>"


def test_nested_NumpyArray_shape():
    data = np.full((3, 3, 5), 999, dtype=np.int64)
    data[1:3] = np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)

    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(data),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::RegularArray_")


def test_nested_RegularArray_NumpyArray():
    ak_array_one = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 999, 999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
            ),
            3,
        ),
    )
    data_frame_one = ak.to_rdataframe({"one": ak_array_one})
    assert str(data_frame_one.GetColumnType("one")).startswith("awkward::RegularArray_")


def test_nested_RegularArray_NumpyArray_two():
    ak_array_two = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.emptyarray.EmptyArray().to_NumpyArray(np.dtype(np.float64)),
            0,
            zeros_length=11,
        ),
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert str(data_frame_two.GetColumnType("two")).startswith("awkward::RegularArray_")


def test_nested_ListArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 4], dtype=np.int64)),
        ak.contents.listarray.ListArray(
            ak.index.Index(np.array([999, 4, 100, 1], np.int64)),
            ak.index.Index(np.array([999, 7, 100, 3, 200], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
            ),
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::ListArray_")


def test_nested_ListOffsetArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.listoffsetarray.ListOffsetArray(
            ak.index.Index(np.array([1, 1, 4, 4, 6, 7], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
            ),
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::ListArray_")


def test_nested_RecordArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            ["x", "y"],
            parameters={"__record__": "Something"},
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith(
        "awkward::RecordArray_Something_"
    )


def test_nested_RecordArray_NumpyArray_two():
    ak_array_two = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            None,
        ),
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert str(data_frame_two.GetColumnType("two")).startswith("awkward::RecordArray_")


def test_nested_RecordArray_NumpyArray_three():
    ak_array_three = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], [], 11),
    )
    data_frame_three = ak.to_rdataframe({"three": ak_array_three})
    assert str(data_frame_three.GetColumnType("three")).startswith(
        "awkward::RecordArray_"
    )


def test_nested_RecordArray_NumpyArray_four():
    ak_array_four = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], None, 11),
    )
    data_frame_four = ak.to_rdataframe({"four": ak_array_four})
    assert str(data_frame_four.GetColumnType("four")).startswith(
        "awkward::RecordArray_"
    )


def test_nested_IndexedArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedarray.IndexedArray(
            ak.index.Index(np.array([999, 2, 2, 0, 1, 4, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::IndexedArray_")


def test_nested_IndexedOptionArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedoptionarray.IndexedOptionArray(
            ak.index.Index(np.array([999, 2, 2, -1, 1, -1, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::IndexedOptionArray_")


def test_nested_ByteMaskedArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 1, 0, 1, 0, 1], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=True,
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::ByteMaskedArray_")


def test_nested_ByteMaskedArray_NumpyArray_two():
    ak_array_two = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 0, 1, 0, 1, 0], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=False,
        ),
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert str(data_frame_two.GetColumnType("two")).startswith(
        "awkward::ByteMaskedArray_"
    )


def test_nested_BitMaskedArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=False,
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::BitMaskedArray_")


def test_nested_BitMaskedArray_NumpyArray_two():
    ak_array_two = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=False,
        ),
    )
    data_frame_two = ak.to_rdataframe({"two": ak_array_two})
    assert str(data_frame_two.GetColumnType("two")).startswith(
        "awkward::BitMaskedArray_"
    )


def test_nested_BitMaskedArray_NumpyArray_three():
    ak_array_three = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=True,
        ),
    )
    data_frame_three = ak.to_rdataframe({"three": ak_array_three})
    assert str(data_frame_three.GetColumnType("three")).startswith(
        "awkward::BitMaskedArray_"
    )


def test_nested_BitMaskedArray_NumpyArray_four():
    ak_array_four = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=True,
        ),
    )
    data_frame_four = ak.to_rdataframe({"four": ak_array_four})
    assert str(data_frame_four.GetColumnType("four")).startswith(
        "awkward::BitMaskedArray_"
    )


def test_nested_UnmaskedArray_NumpyArray():
    ak_array_in = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.unmaskedarray.UnmaskedArray(
            ak.contents.numpyarray.NumpyArray(np.array([999, 0.0, 1.1, 2.2, 3.3]))
        ),
    )
    data_frame = ak.to_rdataframe({"x": ak_array_in})
    assert str(data_frame.GetColumnType("x")).startswith("awkward::UnmaskedArray_")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import os

import numpy as np  # noqa: F401
import pytest

import awkward as ak

pytest.importorskip("pyarrow")
pytest.importorskip("pyarrow.parquet")
pytest.importorskip("fsspec")


def test(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")

    original = ak.Record({"x": 1, "y": [1, 2, 3], "z": "THREE"})

    assert ak.from_arrow(ak.to_arrow(original)).to_list() == original.to_list()

    assert ak.from_arrow(ak.to_arrow_table(original)).to_list() == original.to_list()

    ak.to_parquet(original, filename)
    assert ak.from_parquet(filename).to_list() == original.to_list()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_reduce_1d_strided():
    non_contiguous_array = np.arange(64, dtype=np.int64)[::3]
    layout = ak.contents.NumpyArray(
        non_contiguous_array,
    )
    assert not layout.is_contiguous
    assert ak.sum(layout, axis=-1) == np.sum(non_contiguous_array, axis=-1)


def test_reduce_transpose_2d():
    non_contiguous_array = np.arange(6 * 8, dtype=np.int64).reshape(6, 8).T
    layout = ak.contents.NumpyArray(
        non_contiguous_array,
    )
    assert not layout.is_contiguous
    assert (
        ak.sum(layout, axis=-1).to_list()
        == np.sum(non_contiguous_array, axis=-1).tolist()
    )


def test_reduce_2d():
    non_contiguous_array = np.arange(6 * 8, dtype=np.int64).reshape(6, 8)
    layout = ak.contents.NumpyArray(
        non_contiguous_array,
    )
    assert layout.is_contiguous
    assert (
        ak.sum(layout, axis=-1).to_list()
        == np.sum(non_contiguous_array, axis=-1).tolist()
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

__all__ = ("from_buffers",)
import math

import awkward as ak
from awkward._backends.dispatch import regularize_backend
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import is_integer

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@high_level_function
def from_buffers(
    form,
    length,
    container,
    buffer_key="{form_key}-{attribute}",
    *,
    backend="cpu",
    byteorder="<",
    highlevel=True,
    behavior=None,
):
    """
    Args:
        form (#ak.forms.Form or str/dict equivalent): The form of the Awkward
            Array to _reconstitute from named buffers.
        length (int): Length of the array. (The output of this function is always
            single-partition.)
        container (Mapping, such as dict): The str \u2192 Python buffers that
            represent the decomposed Awkward Array. This `container` is only
            assumed to have a `__getitem__` method that accepts strings as keys.
        buffer_key (str or callable): Python format string containing
            `"{form_key}"` and/or `"{attribute}"` or a function that takes these
            as keyword arguments and returns a string to use as a key for a buffer
            in the `container`.
        backend (str): Library to use to generate values that are
            put into the new array. The default, cpu, makes NumPy
            arrays, which are in main memory (e.g. not GPU). If all the values in
            `container` have the same `backend` as this, they won't be copied.
        byteorder (`"<"`, `">"`): Endianness of buffers read from `container`.
            If the byteorder does not match the current system byteorder, the
            arrays will be copied.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Reconstitutes an Awkward Array from a Form, length, and a collection of memory
    buffers, so that data can be losslessly read from file formats and storage
    devices that only map names to binary blobs (such as a filesystem directory).

    The first three arguments of this function are the return values of
    #ak.to_buffers, so a full round-trip is

        >>> reconstituted = ak.from_buffers(*ak.to_buffers(original))

    The `container` argument lets you specify your own Mapping, which might be
    an interface to some storage format or device (e.g. h5py). It's okay if
    the `container` dropped NumPy's `dtype` and `shape` information, leaving
    raw bytes, since `dtype` and `shape` can be reconstituted from the
    #ak.forms.NumpyForm.
    If the values of `container` are recognised as arrays by the given backend,
    a view over their existing data will be used, where possible.

    The `buffer_key` should be the same as the one used in #ak.to_buffers.

    See #ak.to_buffers for examples.
    """
    return _impl(
        form,
        length,
        container,
        buffer_key,
        backend,
        byteorder,
        highlevel,
        behavior,
        False,
    )


def _impl(
    form,
    length,
    container,
    buffer_key,
    backend,
    byteorder,
    highlevel,
    behavior,
    simplify,
):
    backend = regularize_backend(backend)

    if isinstance(form, str):
        if ak.types.numpytype.is_primitive(form):
            form = ak.forms.NumpyForm(form)
        else:
            form = ak.forms.from_json(form)
    elif isinstance(form, dict):
        form = ak.forms.from_dict(form)

    if not (is_integer(length) and length >= 0):
        raise TypeError("'length' argument must be a non-negative integer")

    if not isinstance(form, ak.forms.Form):
        raise TypeError(
            "'form' argument must be a Form or its Python dict/JSON string representation"
        )

    if isinstance(buffer_key, str):

        def getkey(form, attribute):
            return buffer_key.format(form_key=form.form_key, attribute=attribute)

    elif callable(buffer_key):

        def getkey(form, attribute):
            return buffer_key(form_key=form.form_key, attribute=attribute, form=form)

    else:
        raise TypeError(
            f"buffer_key must be a string or a callable, not {type(buffer_key)}"
        )

    out = _reconstitute(form, length, container, getkey, backend, byteorder, simplify)
    return wrap_layout(out, behavior, highlevel)


_index_to_dtype = {
    "i8": np.dtype("<i1"),
    "u8": np.dtype("<u1"),
    "i32": np.dtype("<i4"),
    "u32": np.dtype("<u4"),
    "i64": np.dtype("<i8"),
}


def _from_buffer(nplike, buffer, dtype, count, byteorder):
    if nplike.is_own_array(buffer):
        array = nplike.reshape(buffer.view(dtype), shape=(-1,), copy=False)

        # Require 1D
        if array.size < count:
            raise TypeError(
                f"size of array ({array.size}) is less than size of form ({count})"
            )

        return array[:count]
    else:
        array = nplike.frombuffer(buffer, dtype=dtype, count=count)
        if byteorder != ak._util.native_byteorder:
            return array.byteswap(inplace=False)
        else:
            return array


def _reconstitute(form, length, container, getkey, backend, byteorder, simplify):
    if isinstance(form, ak.forms.EmptyForm):
        if length != 0:
            raise ValueError(f"EmptyForm node, but the expected length is {length}")
        return ak.contents.EmptyArray()

    elif isinstance(form, ak.forms.NumpyForm):
        dtype = ak.types.numpytype.primitive_to_dtype(form.primitive)
        raw_array = container[getkey(form, "data")]
        real_length = length
        for x in form.inner_shape:
            real_length *= x
        data = _from_buffer(
            backend.nplike,
            raw_array,
            dtype=dtype,
            count=real_length,
            byteorder=byteorder,
        )
        if form.inner_shape != ():
            if len(data) == 0:
                data = backend.nplike.reshape(data, (length, *form.inner_shape))
            else:
                data = backend.nplike.reshape(data, (-1, *form.inner_shape))
        return ak.contents.NumpyArray(
            data, parameters=form._parameters, backend=backend
        )

    elif isinstance(form, ak.forms.UnmaskedForm):
        content = _reconstitute(
            form.content, length, container, getkey, backend, byteorder, simplify
        )
        if simplify:
            make = ak.contents.UnmaskedArray.simplified
        else:
            make = ak.contents.UnmaskedArray
        return make(content, parameters=form._parameters)

    elif isinstance(form, ak.forms.BitMaskedForm):
        raw_array = container[getkey(form, "mask")]
        excess_length = int(math.ceil(length / 8.0))
        mask = _from_buffer(
            backend.index_nplike,
            raw_array,
            dtype=_index_to_dtype[form.mask],
            count=excess_length,
            byteorder=byteorder,
        )
        content = _reconstitute(
            form.content, length, container, getkey, backend, byteorder, simplify
        )
        if simplify:
            make = ak.contents.BitMaskedArray.simplified
        else:
            make = ak.contents.BitMaskedArray
        return make(
            ak.index.Index(mask),
            content,
            form.valid_when,
            length,
            form.lsb_order,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ByteMaskedForm):
        raw_array = container[getkey(form, "mask")]
        mask = _from_buffer(
            backend.index_nplike,
            raw_array,
            dtype=_index_to_dtype[form.mask],
            count=length,
            byteorder=byteorder,
        )
        content = _reconstitute(
            form.content, length, container, getkey, backend, byteorder, simplify
        )
        if simplify:
            make = ak.contents.ByteMaskedArray.simplified
        else:
            make = ak.contents.ByteMaskedArray
        return make(
            ak.index.Index(mask),
            content,
            form.valid_when,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.IndexedOptionForm):
        raw_array = container[getkey(form, "index")]
        index = _from_buffer(
            backend.index_nplike,
            raw_array,
            dtype=_index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
        )
        next_length = (
            0 if len(index) == 0 else max(0, backend.index_nplike.max(index) + 1)
        )
        content = _reconstitute(
            form.content, next_length, container, getkey, backend, byteorder, simplify
        )
        if simplify:
            make = ak.contents.IndexedOptionArray.simplified
        else:
            make = ak.contents.IndexedOptionArray
        return make(
            ak.index.Index(index),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.IndexedForm):
        raw_array = container[getkey(form, "index")]
        index = _from_buffer(
            backend.index_nplike,
            raw_array,
            dtype=_index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
        )
        next_length = (
            0
            if len(index) == 0
            else backend.index_nplike.index_as_shape_item(
                backend.index_nplike.max(index) + 1
            )
        )
        content = _reconstitute(
            form.content, next_length, container, getkey, backend, byteorder, simplify
        )
        if simplify:
            make = ak.contents.IndexedArray.simplified
        else:
            make = ak.contents.IndexedArray
        return make(
            ak.index.Index(index),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ListForm):
        raw_array1 = container[getkey(form, "starts")]
        raw_array2 = container[getkey(form, "stops")]
        starts = _from_buffer(
            backend.index_nplike,
            raw_array1,
            dtype=_index_to_dtype[form.starts],
            count=length,
            byteorder=byteorder,
        )
        stops = _from_buffer(
            backend.index_nplike,
            raw_array2,
            dtype=_index_to_dtype[form.stops],
            count=length,
            byteorder=byteorder,
        )
        reduced_stops = stops[starts != stops]
        next_length = 0 if len(starts) == 0 else backend.index_nplike.max(reduced_stops)
        content = _reconstitute(
            form.content, next_length, container, getkey, backend, byteorder, simplify
        )
        return ak.contents.ListArray(
            ak.index.Index(starts),
            ak.index.Index(stops),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.ListOffsetForm):
        raw_array = container[getkey(form, "offsets")]
        offsets = _from_buffer(
            backend.index_nplike,
            raw_array,
            dtype=_index_to_dtype[form.offsets],
            count=length + 1,
            byteorder=byteorder,
        )
        next_length = 0 if len(offsets) == 1 else offsets[-1]
        content = _reconstitute(
            form.content, next_length, container, getkey, backend, byteorder, simplify
        )
        return ak.contents.ListOffsetArray(
            ak.index.Index(offsets),
            content,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.RegularForm):
        next_length = length * form.size
        content = _reconstitute(
            form.content, next_length, container, getkey, backend, byteorder, simplify
        )
        return ak.contents.RegularArray(
            content,
            form.size,
            length,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.RecordForm):
        contents = [
            _reconstitute(
                content, length, container, getkey, backend, byteorder, simplify
            )
            for content in form.contents
        ]
        return ak.contents.RecordArray(
            contents,
            None if form.is_tuple else form.fields,
            length,
            parameters=form._parameters,
        )

    elif isinstance(form, ak.forms.UnionForm):
        raw_array1 = container[getkey(form, "tags")]
        raw_array2 = container[getkey(form, "index")]
        tags = _from_buffer(
            backend.index_nplike,
            raw_array1,
            dtype=_index_to_dtype[form.tags],
            count=length,
            byteorder=byteorder,
        )
        index = _from_buffer(
            backend.index_nplike,
            raw_array2,
            dtype=_index_to_dtype[form.index],
            count=length,
            byteorder=byteorder,
        )
        lengths = []
        for tag in range(len(form.contents)):
            selected_index = index[tags == tag]
            if len(selected_index) == 0:
                lengths.append(0)
            else:
                lengths.append(backend.index_nplike.max(selected_index) + 1)
        contents = [
            _reconstitute(
                content, lengths[i], container, getkey, backend, byteorder, simplify
            )
            for i, content in enumerate(form.contents)
        ]
        if simplify:
            make = ak.contents.UnionArray.simplified
        else:
            make = ak.contents.UnionArray
        return make(
            ak.index.Index(tags),
            ak.index.Index(index),
            contents,
            parameters=form._parameters,
        )

    else:
        raise AssertionError("unexpected form node type: " + str(type(form)))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("to_numpy",)
import awkward as ak
from awkward._backends.numpy import NumpyBackend
from awkward._dispatch import high_level_function


@high_level_function
def to_numpy(array, *, allow_missing=True):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        allow_missing (bool): allow missing (None) values.

    Converts `array` (many types supported, including all Awkward Arrays and
    Records) into a NumPy array, if possible.

    If the data are numerical and regular (nested lists have equal lengths
    in each dimension, as described by the #ak.Array.type), they can be losslessly
    converted to a NumPy array and this function returns without an error.

    Otherwise, the function raises an error. It does not create a NumPy
    array with dtype `"O"` for `np.object_` (see the
    [note on object_ type](https://docs.scipy.org/doc/numpy/reference/arrays.scalars.html#arrays-scalars-built-in))
    since silent conversions to dtype `"O"` arrays would not only be a
    significant performance hit, but would also break functionality, since
    nested lists in a NumPy `"O"` array are severed from the array and
    cannot be sliced as dimensions.

    If `array` is not an Awkward Array, then this function is equivalent
    to calling `np.asarray` on it.

    If `allow_missing` is True; NumPy
    [masked arrays](https://docs.scipy.org/doc/numpy/reference/maskedarray.html)
    are a possible result; otherwise, missing values (None) cause this
    function to raise an error.

    See also #ak.from_numpy and #ak.to_cupy.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, allow_missing)


def _impl(array, allow_missing):
    import numpy  # noqa: TID251

    with numpy.errstate(invalid="ignore"):
        layout = ak.to_layout(array, allow_record=False)

        backend = NumpyBackend.instance()
        numpy_layout = layout.to_backend(backend)

        return numpy_layout.to_backend_array(allow_missing=allow_missing)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

from awkward._nplikes.typetracer import TypeTracer

typetracer = TypeTracer.instance()


def test_nd():
    data = [
        [
            [1, 2, 3],
            [4, 5, 6],
        ],
        [
            [1, 2, 3],
            [4, 5, 8],
        ],
    ]
    result = typetracer.asarray(data, dtype=np.uint8)
    assert result.dtype == np.dtype(np.uint8)
    assert result.shape == (2, 2, 3)

    # Check default size of array
    array = np.array([1, 2, 3])
    default_int_dtype = array.dtype

    result = typetracer.asarray(data)
    assert result.dtype == default_int_dtype
    assert result.shape == (2, 2, 3)


def test_nd_ragged():
    data = [
        [
            [1, 2, 3],
            [4, 5, 6],
        ],
        [
            [1, 2, 3],
            [4, 5],
        ],
    ]
    with pytest.raises(ValueError, match=r"sequence at dimension .* does not match"):
        typetracer.asarray(data)


def test_unknown_scalar():
    unknown_array = typetracer.asarray([0, 1, 2.0], dtype=np.dtype(np.float64))

    data = [
        [
            [1, 2, 3],
            [4, 5, 6],
        ],
        [
            [1, 2, 3],
            [4, 5, unknown_array[0]],
        ],
    ]
    result = typetracer.asarray(data)
    assert result.dtype == np.dtype(np.float64)
    assert result.shape == (2, 2, 3)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak


def test():
    assert ak.Record({"x": "hello"})["x"] == "hello"
    assert ak.Record({"x": b"hello"})["x"] == b"hello"

    assert ak.Array([{"x": "hello"}])[0, "x"] == "hello"
    assert ak.Array([{"x": b"hello"}])[0, "x"] == b"hello"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import sys

import awkward as ak


class ScalarType:
    def __init__(self, content, behavior=None):
        if not isinstance(content, ak.types.Type):
            raise TypeError(
                "{} all 'contents' must be Type subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        self._content = content
        self._behavior = behavior

    @property
    def content(self):
        return self._content

    @property
    def behavior(self):
        return self._behavior

    def __str__(self):
        return "".join(self._str("", True))

    def show(self, stream=sys.stdout):
        stream.write("".join([*self._str("", False), "\n"]))

    def _str(self, indent, compact):
        return self._content._str(
            indent,
            compact,
            self._behavior,
        )

    def __repr__(self):
        return f"{type(self).__name__}({self._content!r}, {self._behavior!r})"

    def is_equal_to(self, other, *, all_parameters: bool = False) -> bool:
        return isinstance(other, type(self)) and self._content.is_equal_to(
            other._content, all_parameters=all_parameters
        )

    __eq__ = is_equal_to



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from collections.abc import MutableMapping, Sequence

import awkward as ak
from awkward._backends.backend import Backend
from awkward._backends.numpy import NumpyBackend
from awkward._backends.typetracer import TypeTracerBackend
from awkward._errors import AxisError
from awkward._layout import maybe_posaxis
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import ArrayLike, IndexType, NumpyMetadata
from awkward._nplikes.shape import ShapeItem
from awkward._regularize import is_integer_like
from awkward._slicing import NO_HEAD
from awkward._typing import TYPE_CHECKING, Callable, Final, Self, SupportsIndex, final
from awkward._util import UNSET
from awkward.contents.content import Content
from awkward.forms.emptyform import EmptyForm
from awkward.forms.form import Form
from awkward.index import Index

if TYPE_CHECKING:
    from awkward._slicing import SliceItem

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@final
class EmptyArray(Content):
    """
    An EmptyArray is used whenever an array's type is not known because it is empty
    (such as data from #ak.ArrayBuilder without enough sample points to resolve the
    type).

    Unlike all other Content subclasses, EmptyArray cannot contain any parameters
    (parameter values are always None).

    EmptyArray has no equivalent in Apache Arrow.

    To illustrate how the constructor arguments are interpreted, the following is a
    simplified implementation of `__init__`, `__len__`, and `__getitem__`:

        class EmptyArray(Content):
            def __init__(self):
                pass

            def __len__(self):
                return 0

            def __getitem__(self, where):
                if isinstance(where, int):
                    assert False

                elif isinstance(where, slice) and where.step is None:
                    return EmptyArray()

                elif isinstance(where, str):
                    raise ValueError("field " + repr(where) + " not found")

                else:
                    raise AssertionError(where)
    """

    is_unknown = True
    is_leaf = True

    def __init__(self, *, parameters=None, backend=None):
        if not (parameters is None or len(parameters) == 0):
            raise TypeError(f"{type(self).__name__} cannot contain parameters")
        if backend is None:
            backend = NumpyBackend.instance()
        self._init(parameters, backend)

    form_cls: Final = EmptyForm

    def copy(
        self,
        *,
        parameters=UNSET,
        backend=UNSET,
    ):
        if not (parameters is UNSET or parameters is None or len(parameters) == 0):
            raise TypeError(f"{type(self).__name__} cannot contain parameters")
        return EmptyArray(
            backend=self._backend if backend is UNSET else backend,
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    @classmethod
    def simplified(cls, *, parameters=None, backend=None):
        if not (parameters is None or len(parameters) == 0):
            raise TypeError(f"{cls.__name__} cannot contain parameters")
        return cls(backend=backend)

    def _form_with_key(self, getkey: Callable[[Content], str | None]) -> EmptyForm:
        return self.form_cls(form_key=getkey(self))

    def _to_buffers(
        self,
        form: Form,
        getkey: Callable[[Content, Form, str], str],
        container: MutableMapping[str, ArrayLike],
        backend: Backend,
        byteorder: str,
    ):
        assert isinstance(form, self.form_cls)

    def _to_typetracer(self, forget_length: bool) -> Self:
        return EmptyArray(
            backend=TypeTracerBackend.instance(),
        )

    def _touch_data(self, recursive: bool):
        pass

    def _touch_shape(self, recursive: bool):
        pass

    @property
    def length(self) -> ShapeItem:
        return 0

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        extra = self._repr_extra(indent + "    ")
        if len(extra) == 0:
            return indent + pre + "<EmptyArray len='0'/>" + post
        else:
            out = [indent, pre, "<EmptyArray len='0'>"]
            out.extend(extra)
            out.append("\n" + indent + "</EmptyArray>")
            out.append(post)
            return "".join(out)

    def to_NumpyArray(self, dtype, backend=None):
        backend = backend or self._backend
        return ak.contents.NumpyArray(
            backend.nplike.empty(0, dtype=dtype),
            parameters=self._parameters,
            backend=backend,
        )

    def __array__(self, **kwargs):
        return numpy.empty(0, dtype=np.float64)

    def __iter__(self):
        return iter([])

    def _getitem_nothing(self):
        return self

    def _getitem_at(self, where: IndexType):
        raise ak._errors.index_error(self, where, "array is empty")

    def _getitem_range(self, start: SupportsIndex, stop: IndexType) -> Content:
        return self

    def _getitem_field(
        self, where: str | SupportsIndex, only_fields: tuple[str, ...] = ()
    ) -> Content:
        raise ak._errors.index_error(self, where, "not an array of records")

    def _getitem_fields(
        self, where: list[str | SupportsIndex], only_fields: tuple[str, ...] = ()
    ) -> Content:
        if len(where) == 0:
            return self._getitem_range(0, 0)
        raise ak._errors.index_error(self, where, "not an array of records")

    def _carry(self, carry: Index, allow_lazy: bool) -> EmptyArray:
        assert isinstance(carry, ak.index.Index)

        if not carry.nplike.known_data or carry.length == 0:
            return self
        else:
            raise ak._errors.index_error(self, carry.data, "array is empty")

    def _getitem_next_jagged(
        self, slicestarts: Index, slicestops: Index, slicecontent: Content, tail
    ) -> Content:
        raise ak._errors.index_error(
            self,
            ak.contents.ListArray(
                slicestarts, slicestops, slicecontent, parameters=None
            ),
            "too many jagged slice dimensions for array",
        )

    def _getitem_next(
        self,
        head: SliceItem | tuple,
        tail: tuple[SliceItem, ...],
        advanced: Index | None,
    ) -> Content:
        if head is NO_HEAD:
            return self

        elif is_integer_like(head):
            raise ak._errors.index_error(self, head, "array is empty")

        elif isinstance(head, slice):
            raise ak._errors.index_error(self, head, "array is empty")

        elif isinstance(head, str):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak.index.Index64):
            if not head.nplike.known_data or head.length == 0:
                return self
            else:
                raise ak._errors.index_error(self, head.data, "array is empty")

        elif isinstance(head, ak.contents.ListOffsetArray):
            raise ak._errors.index_error(self, head, "array is empty")

        elif isinstance(head, ak.contents.IndexedOptionArray):
            raise ak._errors.index_error(self, head, "array is empty")

        else:
            raise AssertionError(repr(head))

    def _offsets_and_flattened(self, axis: int, depth: int) -> tuple[Index, Content]:
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            raise AxisError(self, "axis=0 not allowed for flatten")
        else:
            offsets = ak.index.Index64.zeros(1, nplike=self._backend.index_nplike)
            return (
                offsets,
                EmptyArray(backend=self._backend),
            )

    def _mergeable_next(self, other: Content, mergebool: bool) -> bool:
        return True

    def _mergemany(self, others: Sequence[Content]) -> Content:
        if len(others) == 0:
            return self
        elif len(others) == 1:
            return others[0]
        else:
            return others[0]._mergemany(others[1:])

    def _fill_none(self, value: Content) -> Content:
        return EmptyArray(backend=self._backend)

    def _local_index(self, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return ak.contents.NumpyArray(
                self._backend.nplike.empty(0, dtype=np.int64),
                parameters=None,
                backend=self._backend,
            )
        else:
            raise AxisError(f"axis={axis} exceeds the depth of this array ({depth})")

    def _numbers_to_type(self, name, including_unknown):
        if including_unknown:
            return self.to_NumpyArray(ak.types.numpytype.primitive_to_dtype(name))
        else:
            return self

    def _is_unique(self, negaxis, starts, parents, outlength):
        return True

    def _unique(self, negaxis, starts, parents, outlength):
        return self

    def _argsort_next(
        self, negaxis, starts, shifts, parents, outlength, ascending, stable
    ):
        as_numpy = self.to_NumpyArray(np.float64)
        return as_numpy._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable
        )

    def _sort_next(self, negaxis, starts, parents, outlength, ascending, stable):
        return self

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return ak.contents.EmptyArray(backend=self._backend)

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
        behavior,
    ):
        as_numpy = self.to_NumpyArray(reducer.preferred_dtype)
        return as_numpy._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
            behavior,
        )

    def _validity_error(self, path):
        return ""

    def _nbytes_part(self):
        return 0

    def _pad_none(self, target, axis, depth, clip):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 != depth:
            raise AxisError(f"axis={axis} exceeds the depth of this array ({depth})")
        else:
            return self._pad_none_axis0(target, True)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        if options["emptyarray_to"] is None:
            return pyarrow.Array.from_buffers(
                ak._connect.pyarrow.to_awkwardarrow_type(
                    pyarrow.null(),
                    options["extensionarray"],
                    options["record_is_scalar"],
                    mask_node,
                    self,
                ),
                length,
                [
                    ak._connect.pyarrow.to_validbits(validbytes),
                ],
                null_count=length,
            )

        else:
            dtype = np.dtype(options["emptyarray_to"])
            next = ak.contents.NumpyArray(
                numpy.empty(length, dtype=dtype),
                parameters=self._parameters,
                backend=self._backend,
            )
            return next._to_arrow(pyarrow, mask_node, validbytes, length, options)

    def _to_backend_array(self, allow_missing, backend):
        return backend.nplike.empty(0, dtype=np.float64)

    def _remove_structure(self, backend, options):
        return [self]

    def _recursively_apply(
        self, action, behavior, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                if options["keep_parameters"]:
                    return self
                else:
                    return EmptyArray(backend=self._backend)

        else:

            def continuation():
                pass

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            behavior=behavior,
            backend=self._backend,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def to_packed(self) -> Self:
        return self

    def _to_list(self, behavior, json_conversions):
        if not self._backend.nplike.known_data:
            raise TypeError("cannot convert typetracer arrays to Python lists")
        return []

    def _to_backend(self, backend: Backend) -> Self:
        return EmptyArray(backend=backend)

    def _is_equal_to(self, other, index_dtype, numpyarray):
        return True
